# Copyright 2025 Sear Gamemode
import seaborn as sns
import matplotlib.pyplot as plt
import mne 
import pandas as pd 
import os 
import numpy as np 
from matplotlib.gridspec import GridSpec
from matplotlib.colors import ListedColormap
from matplotlib.lines import Line2D


def get_clusters(clusters_stats, ch_names, ch_pos, correlation_values):
    figs = []
    data = np.full(len(ch_names), np.nan)
    mask = np.zeros(len(ch_names), dtype=bool)
    ch_name_to_idx = {ch: idx for idx, ch in enumerate(ch_names)}

    clusters = [cluster_stat['channels'] for cluster_stat in clusters_stats]
    
    for cluster_idx, cluster in enumerate(clusters):
        cluster_value = cluster_idx + 1
        for ch_name in cluster:
            if ch_name in ch_name_to_idx:
                idx = ch_name_to_idx[ch_name]
                data[idx] = cluster_value
                mask[idx] = True

    N_clusters = len(clusters)
    cmap = plt.get_cmap('tab20', N_clusters)
    cluster_colors = cmap(np.arange(N_clusters) % 20)
    cluster_cmap = ListedColormap(cluster_colors)

    fig, ax = plt.subplots(figsize=(8, 8))
    im, cn = mne.viz.plot_topomap(
        data,
        pos=ch_pos,
        axes=ax,
        cmap=cluster_cmap,
        names=ch_names,
        show=False,
        vlim=(0.5, N_clusters + 0.5),
        contours=0,
        image_interp='nearest',
        extrapolate='head',
        outlines='head',
        sphere=(0., 0., 0., 0.095)
    )

    legend_elements = []
    for idx, (cluster_stat) in enumerate(clusters_stats):
        avg_corr = cluster_stat['avg_corr']
        max_corr = cluster_stat['max_corr']
        min_corr = cluster_stat['min_corr']
        color = cluster_colors[idx % 20]
        label = f'Cluster {idx+1}: Correlation (avg={avg_corr:.3f}, max={max_corr:.3f}, min={min_corr:.3f})'
        legend_elements.append(Line2D(
            [0], [0],
            marker='s',
            color='w',
            label=label,
            markerfacecolor=color,
            markersize=10
        ))

    ax.set_title('Bridged Clusters')
    ax.legend(
        handles=legend_elements,
        bbox_to_anchor=(1.05, 1),
        loc='upper left',
        borderaxespad=0.,
        fontsize=8
    )
    figs.append(fig)
    plt.close(fig)

    fig = plt.figure(figsize=(8, 6))
    plt.hist(np.abs(correlation_values), bins=100, edgecolor='black')
    plt.title('Histogram of Pairwise Electrode Correlations')
    plt.xlabel('Correlation Coefficient')
    plt.ylabel('Number of Electrode Pairs')
    figs.append(fig)
    plt.close(fig)
    return figs


def ASSR_erp_plot(ERP_c, ERP_n, noised_c_idx, noised_n_idx, roi, comp, baseline=0.2, save=None, vis=False):
    electrodes_to_plot = roi[comp]
    def get_picks_and_data(ERP, electrodes):
        picks = mne.pick_channels(ERP.epochs.info['ch_names'], include=electrodes)
        epochs = ERP.epochs.pick(picks)
        return epochs 
    
    epochs_c = get_picks_and_data(ERP_c, electrodes_to_plot)
    epochs_n = get_picks_and_data(ERP_n, electrodes_to_plot)

    n_epochs_c = len(epochs_c)
    n_epochs_n = len(epochs_n)

    sf = int(epochs_c.info['sfreq'])
    fig_erp = plt.figure(figsize=(16, 12))
    gs_erp = GridSpec(4, 2, figure=fig_erp, hspace=0.7)
    fig_erp.suptitle(f'Event-Related Potencial (ERP), \nCongruent epochs:{n_epochs_c}/{n_epochs_c+len(noised_c_idx)}, Incongruent epochs:{n_epochs_n}/{n_epochs_n+len(noised_n_idx)}')
    t = epochs_c.times
    for i, electrode in enumerate(electrodes_to_plot):
        row, col = divmod(i, 2)
        ax_erp = fig_erp.add_subplot(gs_erp[row, col])
        data_c = epochs_c.get_data(picks=electrode, verbose=False)*1e6
        data_n = epochs_n.get_data(picks=electrode, verbose=False)*1e6
        mean_data_c = np.mean(data_c, axis=0).squeeze()
        mean_data_n = np.mean(data_n, axis=0).squeeze()
        SE_c = np.std(data_c, axis=0, ddof=1).squeeze()/ np.sqrt(data_c.shape[0])
        SE_n = np.std(data_n, axis=0, ddof=1).squeeze()/ np.sqrt(data_n.shape[0])
        
        ax_erp.plot(t, mean_data_c, label='con', color='blue')
        ax_erp.plot(t, mean_data_n, label='incon', color='red')
        ax_erp.fill_between(t, mean_data_c - SE_c, mean_data_c + SE_c, color='blue', alpha=0.3)
        ax_erp.fill_between(t, mean_data_n - SE_n, mean_data_n + SE_n, color='red', alpha=0.3)
        ax_erp.set_ylim([-20, 20])
        ax_erp.set_title(f'{electrode} electrode',  fontsize=10)
        ax_erp.set_xlabel('Time, sec', fontsize=10)
        ax_erp.set_ylabel('Amplitude, $\mu$V', fontsize=10)
        ax_erp.axhline(0, color='black', linewidth=1)
        ax_erp.axvline(0, color='black', linewidth=1)
        ax_erp.grid(True)
        ax_erp.legend(fontsize=8)
        ax_erp.set_box_aspect(0.2) 

    electrode_c_data = [epochs_c.average(picks=electrode, method='mean').data *1e6 for electrode in electrodes_to_plot]
    electrode_n_data = [epochs_n.average(picks=electrode, method='mean').data *1e6 for electrode in electrodes_to_plot]
    mean_data_c_all = np.mean(electrode_c_data, axis=0).squeeze()
    mean_data_n_all = np.mean(electrode_n_data, axis=0).squeeze()

    ax_erp_avg = fig_erp.add_subplot(gs_erp[3, 0])
    ax_erp_avg.plot(t, mean_data_c_all, label='con', color='blue', linestyle='--')
    ax_erp_avg.plot(t, mean_data_n_all, label='incon', color='red', linestyle='--')
    ax_erp_avg.set_ylim([-20, 20])
    ax_erp_avg.set_title(f'Average over epochs-averaged electrodes', fontsize=10)
    ax_erp_avg.set_xlabel('Time, sec', fontsize=10)
    ax_erp_avg.set_ylabel('Amplitude, $\mu$V', fontsize=10)
    ax_erp_avg.axhline(0, color='black', linewidth=1)
    ax_erp_avg.axvline(0, color='black', linewidth=1)
    ax_erp_avg.grid(True)
    ax_erp_avg.legend(fontsize=8)
    ax_erp_avg.set_box_aspect(0.2)

    epochs_c_data = np.mean(epochs_c.get_data(picks=electrodes_to_plot, verbose=False), axis=1)*1e6
    epochs_n_data = np.mean(epochs_n.get_data(picks=electrodes_to_plot, verbose=False), axis=1)*1e6
    mean_data_c_all_epochs = np.mean(epochs_c_data, axis=0)
    mean_data_n_all_epochs = np.mean(epochs_n_data, axis=0)
    SD_c_all_epochs = np.std(epochs_c_data, axis=0, ddof=1)/ np.sqrt(epochs_c_data.shape[0])
    SD_n_all_epochs = np.std(epochs_n_data, axis=0, ddof=1)/ np.sqrt(epochs_n_data.shape[0])
    t_epochs = np.arange(-baseline, len(mean_data_c_all_epochs) / sf - baseline, 1 / sf)

    ax_erp_avg_epochs = fig_erp.add_subplot(gs_erp[3, 1])
    ax_erp_avg_epochs.plot(t_epochs, mean_data_c_all_epochs, label='con', color='blue', linestyle='--')
    ax_erp_avg_epochs.plot(t_epochs, mean_data_n_all_epochs, label='incon', color='red', linestyle='--')
    ax_erp_avg_epochs.fill_between(t_epochs, mean_data_c_all_epochs - SD_c_all_epochs, mean_data_c_all_epochs + SD_c_all_epochs, color='blue', alpha=0.3)
    ax_erp_avg_epochs.fill_between(t_epochs, mean_data_n_all_epochs - SD_n_all_epochs, mean_data_n_all_epochs + SD_n_all_epochs, color='red', alpha=0.3)
    ax_erp_avg_epochs.set_ylim([-20, 20])
    ax_erp_avg_epochs.set_title(f'Average over electrodes and epochs', fontsize=10)
    ax_erp_avg_epochs.set_xlabel('Time, sec', fontsize=10)
    ax_erp_avg_epochs.set_ylabel('Amplitude, $\mu$V', fontsize=10)
    ax_erp_avg_epochs.axhline(0, color='black', linewidth=1)
    ax_erp_avg_epochs.axvline(0, color='black', linewidth=1)
    ax_erp_avg_epochs.grid(True)
    ax_erp_avg_epochs.legend(fontsize=8)
    ax_erp_avg_epochs.set_box_aspect(0.2)
    plt.show()
    if not vis:
        plt.close(fig_erp)
    if save:
        fig_erp.savefig(save)

def ASSR_psd_plot(ERP_c, ERP_n, noised_c_idx, noised_n_idx, roi, comp, save=None, vis=False):
    electrodes_to_plot = roi[comp]
    def get_picks_and_data(ERP, electrodes):
        picks = mne.pick_channels(ERP.epochs.info['ch_names'], include=electrodes)
        epochs = ERP.epochs.pick(picks)
        return epochs 
    
    epochs_c = get_picks_and_data(ERP_c, electrodes_to_plot)
    epochs_n = get_picks_and_data(ERP_n, electrodes_to_plot)

    sf = int(epochs_c.info['sfreq'])
    fig_psd = plt.figure(figsize=(16, 12))
    gs_psd = GridSpec(4, 2, figure=fig_psd, hspace=0.7)
    fig_psd.suptitle('PSD')
    
    for i, electrode in enumerate(electrodes_to_plot):
        row, col = divmod(i, 2)
        ax_psd = fig_psd.add_subplot(gs_psd[row, col])

        mean_data_c = np.mean(epochs_c.average(picks=electrode, method='mean').data, axis=0)
        mean_data_n = np.mean(epochs_n.average(picks=electrode, method='mean').data, axis=0)
        
        psd_c, freqs_c = mne.time_frequency.psd_array_welch(mean_data_c, sfreq=sf, fmin=0, fmax=40, n_fft=sf, verbose=False)
        psd_n, freqs_n = mne.time_frequency.psd_array_welch(mean_data_n, sfreq=sf, fmin=0, fmax=40, n_fft=sf, verbose=False)

        ax_psd.plot(freqs_c, 10 * np.log10(psd_c), label='con', color='blue')
        ax_psd.plot(freqs_n, 10 * np.log10(psd_n), label='n-con', color='red')
        ax_psd.set_title(f'{electrode} electrode', fontsize=10)
        ax_psd.set_xlabel('Frequency (Hz)', fontsize=10)
        ax_psd.set_ylabel('Power (dB)', fontsize=10)
        #ax_psd.axhline(-100, color='black', linewidth=1)
        ax_psd.axvline(0, color='black', linewidth=1)
        ax_psd.grid(True)
        ax_psd.legend(fontsize=8)
        ax_psd.set_box_aspect(0.2)  # ,      

    electrode_c_data = [epochs_c.average(picks=electrode, method='mean').data *1e6 for electrode in electrodes_to_plot]
    electrode_n_data = [epochs_n.average(picks=electrode, method='mean').data *1e6 for electrode in electrodes_to_plot]
    mean_data_c_all = np.mean(electrode_c_data, axis=0).squeeze()
    mean_data_n_all = np.mean(electrode_n_data, axis=0).squeeze()

    #    PSD
    psd_c_all, freqs_c_all = mne.time_frequency.psd_array_welch(mean_data_c_all, sfreq=sf, fmin=0, fmax=40, n_fft=sf, verbose=False)
    psd_n_all, freqs_n_all = mne.time_frequency.psd_array_welch(mean_data_n_all, sfreq=sf, fmin=0, fmax=40, n_fft=sf, verbose=False)

    ax_psd_avg = fig_psd.add_subplot(gs_psd[3, :])
    ax_psd_avg.plot(freqs_c_all, 10 * np.log10(psd_c_all), label='con', color='blue', linestyle='--')
    ax_psd_avg.plot(freqs_n_all, 10 * np.log10(psd_n_all), label='n-con', color='red', linestyle='--')
    ax_psd_avg.set_title(f'Average over electrodes', fontsize=10)
    ax_psd_avg.set_xlabel('Frequency (Hz)', fontsize=10)
    ax_psd_avg.set_ylabel('Power (dB)', fontsize=10)
    #ax_psd_avg.axhline(-100, color='black', linewidth=1)
    ax_psd_avg.axvline(0, color='black', linewidth=1)
    ax_psd_avg.grid(True)
    ax_psd_avg.legend(fontsize=8)
    ax_psd_avg.set_box_aspect(0.1)
    plt.show()
    if not vis:
        plt.close(fig_psd)
    if save:
        fig_psd.savefig(save)


def snr_plot(snr_matrix, ch_names):
    snr_fig, ax = plt.subplots(figsize=(8, 6))
    ep_ind = np.arange(0, snr_matrix.shape[0], 1)
    ch_ind = np.arange(0, snr_matrix.shape[1], 1)
    xticklabels = np.array(ep_ind)[::max(1, len(ep_ind) // 10)]
    yticklabels = np.array(ch_names)[ch_ind]
    
    sns.heatmap(snr_matrix.T, cmap='magma', xticklabels=xticklabels, yticklabels=yticklabels,
                vmin=-30, vmax=-10, ax=ax)
    
    ax.set_xticks(np.arange(len(xticklabels)))
    ax.set_xticklabels(xticklabels)
    ax.set_title('SNR Heatmap')
    ax.set_xlabel('Epochs')
    ax.set_ylabel('Channels')
    ax.tick_params(axis='y', labelsize=8)
    plt.close(snr_fig)

    snr_values = snr_matrix.flatten()
    hist_fig, ax = plt.subplots(figsize=(8, 6))
    sns.histplot(snr_values, bins=30, color='blue', edgecolor='black', ax=ax, kde=True, stat='density')

    ax.set_title('SNR Value Frequency Histogram')
    ax.set_xlabel('SNR Value')
    ax.set_ylabel('Probability')
    ax.grid(axis='y', alpha=0.75)
    ax.set_xlim(-30, 0)
    plt.close(hist_fig)
    return snr_fig, hist_fig


def compared_snr(snr_matrices, names):
    kde_fig, ax = plt.subplots(figsize=(10, 6))
    colors = sns.color_palette("husl", len(snr_matrices))
    for snr_matrix, name, color in zip(snr_matrices, names, colors):
        snr_values = snr_matrix.flatten()
        mean_snr = np.median(snr_values)
        label_with_mean = f"{name} (Median: {mean_snr:.2f})"
        kde = sns.kdeplot(snr_values, ax=ax, color=color, fill=True, bw_adjust=0.5, label=label_with_mean, alpha=0.5)
        
        color = kde.collections[-1].get_facecolor()[0]  #    RGBA
        ax.axvline(mean_snr, linestyle='--', color=color, linewidth=1.5)
    ax.set_title('SNR KDE')
    ax.set_xlabel('SNR Value')
    ax.set_ylabel('Density')
    ax.grid(axis='y', alpha=0.75)
    ax.set_xlim(-35, 0)
    ax.set_ylim(0, 0.15)
    ax.grid()
    ax.legend(title='SNR Matrices', loc='upper right')
    plt.close(kde_fig)
    return kde_fig


def bridge_save(raw, correlation_matrix, bridged_idx, ed_matrix, saving_dir=None, vis=False):
    channels = raw.ch_names
    bridge_fig = mne.viz.plot_bridged_electrodes(raw.info, bridged_idx, ed_matrix)
    fig_corr = plt.figure(figsize=(8, 6))
    ax = sns.heatmap(correlation_matrix, cmap='viridis', xticklabels=channels, yticklabels=channels) #np.mean(ed_matrix, axis=0)
    ax.set_xticklabels(channels, rotation=90, fontsize=7)
    ax.set_yticklabels(channels, rotation=0, fontsize=7)
    plt.title('Correlation of Channel Connections (R)')
    plt.xlabel('Channels')
    plt.ylabel('Channels')
    if saving_dir:
        plt.savefig(os.path.join(saving_dir, f'orrelation_electrodes.png'))
        bridge_fig.savefig(os.path.join(saving_dir, f'bridged_electrodes.png'))
    if not vis:
        plt.close(fig_corr)
        plt.close(bridge_fig)


def components_save(raw, ica, exclude_idx, labels, probas, saving_dir=None, vis=False):
    duration = raw.times[-1]*1000
    start_time = 0 
    stop_time = min(10*1000, duration)  
    ica_fig = ica.plot_components(title='ICA Components', show=vis)
    if isinstance(raw, mne.io.Raw) or isinstance(raw,mne.io.brainvision.brainvision.RawBrainVision):
        overlay_fig = ica.plot_overlay(raw, title='Overlay', start=start_time, stop=stop_time, 
                                    exclude=exclude_idx, show=vis)
    else:
        overlay_fig = ica.plot_overlay(raw.average(), title='Overlay', start=start_time, stop=stop_time, 
                                    exclude=exclude_idx, show=vis)
    if saving_dir:
        ica_fig.savefig(os.path.join(saving_dir, 'ica_components.png'))
        overlay_fig.savefig(os.path.join(saving_dir, 'overlay.png'))
    if not vis:
        plt.close(ica_fig)
        plt.close(overlay_fig)
    for idx in exclude_idx:
        ica_exclude_fig = ica.plot_properties(raw, picks=idx, show=vis, verbose=False)[0]
        if saving_dir:
            if labels is None:
                ica_exclude_fig.savefig(os.path.join(saving_dir, f'ica_exclude_components_{idx}_label:hand.png'))
            else:
                ica_exclude_fig.savefig(os.path.join(saving_dir, f'ica_exclude_components_{idx}_{labels[idx]}_{probas[idx]}.png'))
        if not vis:
            plt.close(ica_exclude_fig)


def event_saver(proc_counter, mind_counter, saving_dir=None, vis=False):
    df = pd.read_csv('tools/ASSR.csv', index_col=None)
    combined_dict = {**proc_counter, **mind_counter}
    combined_dict = {key.replace('Stimulus/', ''): value for key, value in combined_dict.items()}
    df.columns = df.columns.str.replace(' ', '')
    new_row = {column: combined_dict[column] for column in df.columns}
    new_row_df = pd.DataFrame([new_row])
    df = pd.concat([df, new_row_df], ignore_index=True)
    target_row = df.iloc[0].copy()
    target_row_numeric = pd.to_numeric(target_row, errors='coerce')
    new_row_numeric = pd.to_numeric(pd.Series(new_row), errors='coerce')
    difference_row = target_row_numeric - new_row_numeric
    difference_row_df = pd.DataFrame([difference_row])
    df = pd.concat([df, difference_row_df], ignore_index=True)

    df.insert(0, 'Cond', ['Target', 'Real', 'Diff'])
    if saving_dir:
        df.to_csv(os.path.join(saving_dir, 'stimulus_counts.csv'), index=False)
    fig_table, ax = plt.subplots(figsize=(12, 2))  
    ax.axis('off')
    table = ax.table(cellText=df.values, colLabels=df.columns, cellLoc='center', loc='center')
    table.auto_set_font_size(False)
    table.set_fontsize(10)
    table.scale(1.2, 1.2)
    plt.title('  ', fontsize=14, pad=1)
    if saving_dir:
        fig_table.savefig(os.path.join(saving_dir, 'stimulus_counts.png'), bbox_inches='tight', pad_inches=0.1)
    if not vis:
        plt.close(fig_table)


def tfr_plot(tfr, itc, picks, fmin, fmax, vmin=None, vmax=None, baseline=(None, 0), combine='mean', 
             baseline_mode='logratio', saving_dir=None, vis=False, 
             tfr_title='Event Related Spectral Perturbation (ERSP)',
             itc_title='Inter-Trial Coherence (ITC)'):  
    tfr_fig = tfr.plot(title=tfr_title, fmin=fmin, fmax=fmax, vlim=(vmin,vmax), 
             picks=picks, combine=combine, baseline=baseline, mode=baseline_mode)[0] 
    itc_fig = itc.plot(title=itc_title, fmin=fmin, fmax=fmax, 
             picks=picks, combine=combine, baseline=baseline, mode=baseline_mode)[0]
    if not vis:
        plt.close(tfr_fig)
        plt.close(itc_fig)
    if saving_dir:
        tfr_fig.savefig(os.path.join(saving_dir, f'ERSP.png'))
        itc_fig.savefig(os.path.join(saving_dir, f'ITC.png'))


def plot_noise_data(raw, noisy_channel_flags, n_channels, threshold_noise, integral_curve):
    fig = plt.figure(figsize=(18, 8))
    gs = GridSpec(1, 3, width_ratios=[5, 1, 3], figure=fig)

    ax0 = fig.add_subplot(gs[0, 0])
    sns.heatmap(noisy_channel_flags, vmin=0, vmax=1, cmap='viridis', 
                cbar_kws={'label': '  PSD'}, ax=ax0)
    ax0.set_xlabel('')
    ax0.set_ylabel('')
    ax0.set_title('   PSD    ')
    ax0.set_yticks(np.arange(n_channels) + 0.5)
    ax0.set_yticklabels(raw.ch_names, rotation=0)
    
    ax1 = fig.add_subplot(gs[0, 1], sharey=ax0)
    ax1.plot(integral_curve, np.arange(n_channels)+0.5, color='blue', label='  ')
    ax1.axvline(threshold_noise, color='red', linestyle='--', label=f' - {threshold_noise}')
    ax1.set_xlim((0, 1))
    ax1.set_xlabel(' ')
    ax1.set_title(' ')
    ax1.legend(loc='right', fontsize='small')
    ax1.grid(True)

    ax2 = fig.add_subplot(gs[0, 2])  #    
    pos = np.array([raw.info['chs'][i]['loc'][:2] for i in range(raw.info['nchan'])]) 
    mne.viz.plot_topomap(
        integral_curve, pos, axes=ax2, show=False, cmap='RdBu_r', vlim=(0,1),
        names=raw.ch_names
    )
    ax2.set_title('  ')
    plt.tight_layout()
    plt.show()


def plot_topomap(coherence, band_name, info, title):
    coherence_mean = coherence.mean(axis=1)
    mask = np.zeros((len(info['ch_names']), len(info['ch_names'])), dtype=bool)
    for i in range(len(info['ch_names'])):
        for j in range(i+1, len(info['ch_names'])):
            mask[i, j] = True
    mne.viz.plot_topomap(coherence_mean, info, mask=mask, show=False)
    plt.title(title)
    plt.show()


def char_plot(spec_con, epochs, method, mode, cwt_freqs):
    con = spec_con.get_data()
    if np.iscomplexobj(con):
        con = np.abs(con)
    con_mean = con.mean(axis=0)
    times = epochs.times
    plt.figure(figsize=(12, 6))
    plt.imshow(con_mean, aspect=1/40, cmap='inferno', extent=[times[0], times[-1], cwt_freqs[0], cwt_freqs[-1]], origin='lower') #vmin=0, vmax=1
    plt.colorbar(label=f' {method} {mode}')
    plt.xlabel('Time (s)')
    plt.ylabel('Frequency (Hz)')
    plt.title(f'  {method}   CWT Morlet')
    plt.show()
    plt.savefig(f'{method} {mode}.png')


def plot_circular_connectivity(con, epochs, threshold=0.5, title='Circular Connectivity Diagram'):
    con = np.mean(np.abs(con), axis=2)  
    con[con < threshold] = 0  
    node_names = epochs.ch_names
    mne.viz.plot_connectivity_circle(con, node_names, title=title, vmin=0, vmax=1, colormap='inferno', show=True)



# Copyright 2025 Sear Gamemode
import mne 
import os
import numpy as np
import networkx as nx
from tqdm import tqdm
import matplotlib.pyplot as plt
from collections import defaultdict
from mne.preprocessing import compute_bridged_electrodes
from itertools import chain
from scipy.sparse import csr_matrix
from scipy.sparse.csgraph import connected_components
import json

from autoreject import Ransac 
from joblib import Parallel, delayed
from .savers import bridge_save, event_saver, get_clusters
from .metrics import isolation_forest, check_volt_of_epochs 
from .scenarious import verificate_events, is_subsequence
from .montages import create_custom_montage, read_elc, align_head

def compute_bad_epochs(epochs, snr_matrix, roi_channels=None, thr_auto=True):
    n_trials, n_channels, n_times = epochs._data.shape
    ch_names = epochs.ch_names 
    
    if thr_auto:
        threshold_ep=np.percentile(snr_matrix, 20, axis=(0, 1))
        threshold_ch=np.percentile(snr_matrix, 0.5, axis=(0, 1))
        volt_max=dict(eeg=np.percentile(np.abs(epochs._data), 99.999, axis=(0, 1, 2)))
        volt_min=dict(eeg=np.percentile(np.abs(epochs._data), 10, axis=(0, 1, 2)))
    else:
        threshold_ep=-22
        threshold_ch=-30
        volt_max = dict(eeg=130e-6)
        volt_min = dict(eeg=1e-6)

    #--------------SNR---------------#
    rej_dict = {}
    
    rej_dict['SNR_over_channels'] = np.where(np.mean(snr_matrix, axis=1) < threshold_ep)[0].tolist() #   ,      
    rej_dict['SNR_channel'] = []
    for epoch_idx in range(n_trials):
        channel_names_above_threshold = [ch_names[ch] for ch in np.where(snr_matrix[epoch_idx] < threshold_ch)[0]] #     
        if channel_names_above_threshold:
            rej_dict['SNR_channel'].append((epoch_idx, channel_names_above_threshold)) #         

    #-------ISOLATION_FOREST---------#
    rej_dict['Isol_forest_all_chns'] = isolation_forest(epochs.copy(), mode='ep')
    if roi_channels:
        rej_dict['Isol_forest_tar_chns'] = isolation_forest(epochs.copy().pick(roi_channels), mode='ep')
    #------------VOLT----------------#
    rej_dict['Volt_max'] = check_volt_of_epochs(epochs, reject=volt_max, flat=None)
    rej_dict['Volt_flat'] = check_volt_of_epochs(epochs, reject=None, flat=volt_min)

    #--------------FINAL-------------#
    snr_all_channels_indices = set(rej_dict['SNR_over_channels'])
    snr_channel_indices = set(epoch for epoch, _ in rej_dict['SNR_channel'])
    volt_max_indices = set(rej_dict['Volt_max'])
    volt_min_indices = set(rej_dict['Volt_flat'])
    isol_forest_all_indices = set(rej_dict['Isol_forest_all_chns'])

    combined_indices = snr_all_channels_indices | snr_channel_indices | volt_max_indices | isol_forest_all_indices | volt_min_indices
    if roi_channels:
        isol_forest_tar_indices = set(rej_dict['Isol_forest_tar_chns'])
        combined_indices = combined_indices | isol_forest_tar_indices
        
    rej_dict['FINAL'] = sorted(list(combined_indices))
    rej_dict['Percentage_removed_trials'] = len(rej_dict['FINAL'])/epochs._data.shape[0]*100
    #for key in rej_dict.keys():
    #    print(f'rej_dict[{key}] = {rej_dict[key]}')
    return rej_dict

def set_montage(raw, montage, elc_file, mode, threshold, verbose=False, interpolate=None, vis=None):
    if montage=='waveguard64':
        montage = create_custom_montage(montage)
    elif montage == 'personal':
        ch_dict, nasion, lpa, rpa, hsp = read_elc(elc_file)
        if interpolate:
            ch_dict, nasion, lpa, rpa, hsp = align_head(ch_dict, nasion, np.array(lpa), np.array(rpa), np.array(hsp), standard='waveguard64', 
                                                    mode=mode, threshold=threshold)
        montage = mne.channels.make_dig_montage(ch_pos=ch_dict, nasion=nasion, lpa=lpa, rpa=rpa, hsp=hsp, coord_frame='head')
    else:
        montage = mne.channels.make_standard_montage(montage)
    
    raw.set_montage(montage, verbose=False)

    if vis:
        fig = montage.plot(show_names=True, kind='3d')
        for ax in fig.get_axes():
            if ax.name == '3d':
                ax.set_xlim([-0.1, 0.1])
                ax.set_ylim([-0.1, 0.1])
                ax.set_zlim([-0.1, 0.1])
    return raw


def detect_bad_channels(raw, method='auto'):
    bad_channels = []
    scores = []
    electrodesD = {}
    flat_chans, noisy_channels, _, _ = get_flat_channels(raw)
    clusters, bridge_figs = search_bridge_cluster(raw)
    bridged_electrodes = list(chain.from_iterable(clusters))
    raw = raw.drop_channels(np.concatenate([flat_chans, noisy_channels, bridged_electrodes]))
    if method == 'ransac':
        bad_channels, scores, noised_fig = DNC_ransac(raw)
    elif method == 'ed':
        bad_channels, scores, noised_fig = DNC_electrical_distance(raw)
    elif method == 'corr':
        bad_channels, scores, noised_fig = DNC_corr(raw)
    elif method == 'lof':
        bad_channels, scores = mne.preprocessing.find_bad_channels_lof(raw, metric="euclidean", return_scores=True, verbose=False)
    elif method in ['auto', 'SN_ratio']:
        bad_channels, scores, noised_fig = DNC_SN_ratio(raw)
    else:
        raise ValueError(f"Unknown method '{method}'. Please use 'ransac', 'neighbours', 'psd', 'ed', 'corr' or 'auto'")
    
    electrodesD['HighAmp'] = noisy_channels
    electrodesD['LowAmp'] = flat_chans
    electrodesD['Bridged'] = bridged_electrodes
    electrodesD['Noise_Rate'] = bad_channels
    return electrodesD, clusters, bridge_figs, noised_fig
    

def get_flat_channels(raw, threshold_min=3e-6, threshold_max=300e-6, threshold_length=0.5):
    data = raw.get_data()
    ch_names = np.array(raw.ch_names)
    empty_scores = np.array([np.mean(np.abs(data[idx]) < threshold_min) for idx, ch in enumerate(ch_names)])
    empty_channels = np.array(ch_names[empty_scores > threshold_length])
    max_scores = np.array([np.mean(np.abs(data[idx]) > threshold_max) for idx, ch in enumerate(ch_names)])
    max_channels = ch_names[max_scores > threshold_length] 
    return empty_channels, max_channels, empty_scores, max_scores

def search_bridge_cluster(raw, threshold=0.99):
    data = raw.get_data()
    ch_names = raw.ch_names
    corr_matrix = np.corrcoef(data)

    #heuristics
    corr_matrix = (corr_matrix + corr_matrix.T) / 2
    np.fill_diagonal(corr_matrix, 0)

    binary_matrix = (corr_matrix > threshold).astype(int)
    G = nx.from_numpy_array(binary_matrix) # get graph from matrix
    components = list(nx.connected_components(G)) # calculate connected components
    multi_node_components = [comp for comp in components if len(comp) > 1]
    clusters_indices = [list(comp) for comp in multi_node_components]
    clusters_names = [[ch_names[i] for i in comp] for comp in clusters_indices]
    clusters_stats = []
    for indices, comp_names in zip(clusters_indices, clusters_names):
        corr_vals = []
        for i in range(len(indices)):
            for j in range(i+1, len(indices)):
                corr_vals.append(corr_matrix[indices[i], indices[j]])
        corr_vals = np.array(corr_vals)
        stats = {
            'channels': comp_names,
            'max_corr': np.max(corr_vals),
            'min_corr': np.min(corr_vals),
            'avg_corr': np.mean(corr_vals)
        }
        clusters_stats.append(stats)
    ch_pos = np.array([raw.info['chs'][i]['loc'][:2] for i in range(raw.info['nchan'])])
    figs = get_clusters(clusters_stats, ch_names, ch_pos, corr_matrix.flatten())
    return clusters_names, figs


def search_bridge_cluster_with_times(raw, window_size=2.0, overlap=0.5):
    data = raw.get_data()
    sfreq = raw.info['sfreq']
    ch_names = raw.ch_names
    window_samples = int(window_size * sfreq)
    step_samples = int(window_samples * (1 - overlap))
    clusters_list = []
    bridge_times = []
    for start in range(0, data.shape[1] - window_samples + 1, step_samples):
        end = start + window_samples
        window_data = data[:, start:end]
        clusters, _ = search_bridge_cluster(window_data, ch_names=ch_names, plot=False)
        clusters_list.append(clusters)
        bridge_times.append(start)
    return clusters_list, bridge_times


def event_check(raw, mind_stimulus, proc_stimulus, saving_dir=None, vis=False):
    proc_count = 0
    mind_count = 0
    other_count = 0
    events, event_id = mne.events_from_annotations(raw, verbose=False)
    rev_event_id = dict(list(zip(event_id.values(), event_id.keys())))
    mind_counter = {stimulus: 0 for stimulus in mind_stimulus}
    proc_counter = {stimulus: 0 for stimulus in proc_stimulus}
    strange_list = set()
    for event in events:
        if rev_event_id[event[2]] in proc_counter.keys():
            for key in proc_counter.keys():
                if rev_event_id[event[2]] == key:
                    proc_counter[key]+=1
                    proc_count += 1
        elif rev_event_id[event[2]] in mind_counter.keys():
            for key in mind_counter.keys():
                if rev_event_id[event[2]] == key:
                    mind_counter[key]+=1
                    mind_count += 1
        else:
            strange_list.add(rev_event_id[event[2]])
            other_count += 1
    event_saver(proc_counter, mind_counter, saving_dir, vis)

def bridging_test(raw, saving_dir=None, vis=False):
    raw = raw.copy()
    bridged_idx, ed_matrix = compute_bridged_electrodes(raw, verbose=False)
    correlation_matrix = np.corrcoef(raw.get_data())
    bridge_save(raw, correlation_matrix, bridged_idx, ed_matrix, saving_dir, vis=vis)
    return bridged_idx

def find_adj_neighbors(raw, ch_name):
    montage = raw.get_montage()
    adjacency, ch_names = mne.channels.find_ch_adjacency(raw.info, 'eeg')
    adjacency = adjacency.toarray()
    idx = raw.ch_names.index(ch_name)
    neighbors_idx = np.where(adjacency[idx])[0]
    neighbors = [ch_names[i] for i in neighbors_idx]
    return neighbors

def calculate_correlations(raw, duration=1, overlap=0.5):
    epochs = mne.make_fixed_length_epochs(raw, duration=duration, overlap=overlap, verbose=False)
    data = epochs.get_data()
    n_trials, n_channels, n_times = epochs.get_data().shape

    adjacency, ch_names = mne.channels.find_ch_adjacency(raw.info, ch_type="eeg")

    neighbors_indices = [np.where(adjacency[i])[0] for i in range(n_channels)]
    correlations = np.zeros((n_trials, n_channels))  
    
    for i in range(n_trials):
        epoch_data = data[i] 
        for j in range(n_channels):
            neighbor_data = np.mean(epoch_data[neighbors_indices[j]], axis=0)
            channel_data = epoch_data[j]
            corr = np.corrcoef(channel_data, neighbor_data)[0, 1]
            correlations[i, j] = corr 
    return correlations

def compute_snr(signal, noise, eps=1e-12):
    power_signal = np.mean(signal ** 2)
    power_noise = np.mean(noise ** 2)
    snr_linear = power_signal / (power_noise + eps) # denominator is greater than zero
    snr_db = 10 * np.log10(snr_linear)
    return snr_linear, snr_db

def q_normalization(x, q1=75, q2=25):
    q75 = np.percentile(x, q1, axis=(0, 1), keepdims=True)
    q25 = np.percentile(x, q2, axis=(0, 1), keepdims=True)
    iqr = q75 - q25
    x_norm = x/iqr
    return x_norm 

def sigmoid(x):
    return 1/(1 + np.exp(-x))

def plot_topomap(variable, raw, title, label, vlim):
    """Plot topomap of Variable"""
    montage = raw.get_montage()
    ch_pos_dict = montage.get_positions()['ch_pos']
    ch_pos = np.array([ch_pos_dict[ch_name][:2] for ch_name in raw.ch_names])
    fig, ax = plt.subplots()
    im, _ = mne.viz.plot_topomap(
                variable, ch_pos, axes=ax, show=False,
                names=raw.ch_names, cmap='viridis', vlim=vlim)
    ax.set_title(title)
    cbar = plt.colorbar(im, ax=ax, shrink=0.8)
    cbar.set_label(label)
    return fig



class ENDetector():
    def __init__(self, k=2.0):
        self.model_params = None
        self.k = k

    def load(self):
        base_path = os.path.dirname(os.path.abspath(__file__))
        file_path = os.path.join(base_path, 'model_params.json')
        with open(file_path, 'r') as f:
            self.model_params = json.load(f)
            self.k = self.model_params["k"]
    
    def update(self):
        base_path = os.path.dirname(os.path.abspath(__file__))
        file_path = os.path.join(base_path, 'model_params.json')
        with open(file_path, 'w') as f:
            json.dump(self.model_params, f, indent=2)

    def fit(self, X):
        median_global = np.median(X)
        mad_global = np.median(np.abs(X - median_global))
        global_threshold = median_global - self.k * mad_global
        print(f"dB Threshold: {global_threshold:.2f} dB")
        self.model_params = {
                "median_global": float(median_global),
                "mad_global": float(mad_global),
                "k": self.k,
                "snr_db_ref": X.tolist()}
        print(self.model_params)

    def predict(self, X):
        median_g = self.model_params["median_global"]
        mad_g = self.model_params["mad_global"]
        k_g = self.model_params["k"]
        threshold = median_g - k_g * mad_g
        noisy_mask = (X < threshold)
        noisy_indices = np.where(noisy_mask)[0]
        return noisy_indices, threshold
    
    def predict_proba(self, X):
        sorted_ref = np.sort(self.model_params["snr_db_ref"])
        n = len(sorted_ref)
        
        snr_db_new = np.atleast_1d(np.array(X, dtype=float))
        prob_noises = []
        
        for x in snr_db_new:
            idx = np.searchsorted(sorted_ref, x, side='right')
            cdf_val = idx / n
            prob_noise = 1 - cdf_val
            prob_noises.append(prob_noise)

        return np.array(prob_noises).flatten()



def DNC_SN_ratio(raw, noise_threshold=0.8, optimized=False):  
    ch_names = raw.ch_names
    original_data = raw.get_data()

    def process_channel(ch_idx):
        ch_name = ch_names[ch_idx]
        raw_channel = raw.copy().load_data()
        raw_channel.pick(ch_names)
        raw_channel.info['bads'] = [ch_name]
        noised_data = original_data[ch_idx, :].copy()
        raw_channel.interpolate_bads(reset_bads=False, verbose=False)
        interpolated_data = raw_channel.get_data(picks=[ch_name])[0] # signal
        noise_data = noised_data - interpolated_data # noise
        _, snr_db = compute_snr(interpolated_data, noise_data)
        return ch_name, snr_db

    if optimized:
        results = Parallel(n_jobs=4)(
            delayed(process_channel)(ch_idx) for ch_idx in range(len(ch_names))
        )
    else:
        results = []
        for ch_idx in range(len(ch_names)):
            result = process_channel(ch_idx)
            results.append(result)
    
    snr_probabilities = {}
    bad_channels = []
    scores = []
    detector = ENDetector(k=2)
    detector.load()

    for ch_name, snr_db in results:
        proba = float(detector.predict_proba(snr_db))
        snr_probabilities[ch_name] = proba
        if proba > noise_threshold:
            bad_channels.append(ch_name)
            scores.append(proba)

    variable = np.array(list(snr_probabilities.values())).flatten().tolist()
    title = 'SNR Probaility topomap'
    label = 'Probability Coefficient'
    vlim=(None, None)
    fig = plot_topomap(variable, raw, title, label, vlim)
    noisy_text = "Noisy Electrodes:\n" + "\n".join(
        [f"{ch_name}: {prob:.2f}" for ch_name, prob in zip(bad_channels, scores)]
    )
    fig.text(1.1, 0.5, noisy_text, va='center', ha='left', fontsize=10)
    plt.close(fig)
    return bad_channels, scores, fig


def DNC_corr(raw, threshold_corr=0.65, threshold_perc=5):
    correlations = calculate_correlations(raw)
    variable=np.abs(np.mean(correlations, axis=0))
    title='Correlation topomap'
    label = 'Correlation Coefficient'
    vlim=(0.5, 1)
    fig = plot_topomap(variable, raw, title, label, vlim)
    plt.close(fig)

    result = np.mean(correlations, axis=0)
    p5 = np.percentile(correlations, threshold_perc, axis=(0, 1))
    artefacts_indexes = result < p5

    binary_matrix = correlations < threshold_corr
    probs = binary_matrix.mean(axis=0)
    noised_channels = ((np.array(raw.ch_names)[artefacts_indexes]).tolist())
    score = probs[artefacts_indexes]
    return noised_channels, score, [fig]


def DNC_electrical_distance(raw, vis=False, thr_std=3):
    
    bridged_idx, ed_matrix = compute_bridged_electrodes(raw, verbose=False)
  
    ed_matrix = ed_matrix.copy()
    picks = mne.pick_types(raw.info, eeg=True)
    tril_idx = np.tril_indices(picks.size)
    for epo_idx in range(ed_matrix.shape[0]):
        ed_matrix[epo_idx][tril_idx] = ed_matrix[epo_idx].T[tril_idx]
    channel_names = np.array([raw.ch_names[i] for i in picks])

    ed_matrix = np.nanmin(ed_matrix, axis=1)

    ed_matrix_norm = q_normalization(ed_matrix)
    elec_dists = np.median(ed_matrix_norm, axis=0)
    
    picks = mne.pick_types(raw.info, eeg=True)
    ch_names = raw.ch_names

    results = (ed_matrix_norm - elec_dists).mean(axis=0)
    probs = sigmoid(results)
    artefact_indices = results > thr_std
    variable = probs
    title = "Electrical Distance Probability topomap"
    label = 'Probability Coefficient'
    vlim = (0, 1.0)
    fig = plot_topomap(variable, title, label, vlim)
    plt.close(fig)
    return (channel_names[artefact_indices]).tolist(), probs[artefact_indices], [fig]


def DNC_ransac(raw):  
    ransac = Ransac(verbose=False)
    epochs = mne.make_fixed_length_epochs(raw, duration=10, overlap=0.5, preload=True)
    _ = ransac.fit_transform(epochs)
    probabilities = ransac.bad_log.mean(axis=0)
    bad_channels = ransac.bad_chs_
    scores = probabilities[mne.pick_channels(raw.info['ch_names'], bad_channels)]
    variable = probabilities
    title = "RANSAC Probability topomap"
    label = 'Probability Coefficient'
    vlim = (0, 1.0)
    fig = plot_topomap(variable, title, label, vlim)
    plt.close(fig)
    return bad_channels, scores, fig


def compared_spectrum(inst1, inst2, fmin=0, fmax=100):
    psd_before = inst1.compute_psd(fmin=fmin, fmax=fmax, remove_dc=False, verbose=False)
    psd_after = inst2.compute_psd(fmin=fmin, fmax=fmax, remove_dc=False, verbose=False)

    if isinstance(inst1, mne.Epochs) and isinstance(inst2, mne.Epochs):
        psd_before = psd_before.average()
        psd_after = psd_after.average()

    bands = {'Delta': (0.5, 4), 'Theta': (4, 8), 'Alpha': (8, 13),
            'Beta': (13, 30), 'Gamma': (30, 80)}
    
    mean_psd_before = 10*np.log10(psd_before.get_data() * 1e12)
    mean_psd_after = 10*np.log10(psd_after.get_data() * 1e12)
    freqs = psd_before.freqs
    psd_list = [mean_psd_before, mean_psd_after]
    titles = ['Spectrum before', 'Spectrum after']
    colors = ['b', 'g']

    ymin = np.floor(min(mean_psd_before.min(), mean_psd_after.min()) / 10) * 10
    ymax = np.ceil(max(mean_psd_before.max(), mean_psd_after.max()) / 10) * 10
    fig, axs = plt.subplots(2, 1, figsize=(14, 10), sharex=False)
  
    for idx, (mean_psd, title, color) in enumerate(zip(psd_list, titles, colors)):
        ax = axs[idx]
        y_mean = np.mean(mean_psd, axis=0)
        y_std = np.std(mean_psd, axis=0)
        
        ax.plot(freqs, y_mean, color=color)
        ax.fill_between(freqs, y_mean - y_std, y_mean + y_std, color=color, alpha=0.3)
        ax.set_title(title)
        ax.set_ylabel('PSD (dB/V/Hz)')
        ax.grid(True)
          
        band_powers = {}
        for band_name, (fmin, fmax) in bands.items():
            idx_band = np.logical_and(freqs >= fmin, freqs <= fmax)
            band_power = y_mean[idx_band].mean()
            band_powers[band_name] = band_power
                 
            ax.axvline(fmin, color='red', linestyle='--', linewidth=1)
            ax.axvline(fmax, color='red', linestyle='--', linewidth=1)
            ax.fill_betweenx([ymin, ymax], fmin, fmax, color='grey', alpha=0.05)
              
            ax.text((fmin + fmax) / 2, ymax - 2, f"{band_name}\n{band_power:.1f} dB",
                    horizontalalignment='center', verticalalignment='top', fontsize=9, zorder=4)
        
        ax.set_ylim(ymin, ymax)
        ax.set_xlim(freqs.min(), freqs.max())
        ax.set_xticks(np.arange(freqs.min(), freqs.max()+1, 2))
        ax.set_yticks(np.arange(ymin, ymax+1, 5))
        ax.set_xlabel('Frequency (Hz)')

    plt.tight_layout()
    plt.close(fig)
    return fig



# Copyright 2025 Sear Gamemode
import mne 
import numpy as np 
import pandas as pd
import matplotlib.pyplot as plt
from scipy.stats import shapiro, normaltest, chisquare, norm
from sklearn.ensemble import IsolationForest
from scipy.signal import correlate, welch
from mne.channels import find_ch_adjacency
from .craft_events import make_RiTi_events


def get_neighbors(inst):
    adjacency, ch_names = find_ch_adjacency(info=inst.info, ch_type='eeg')
    adjacency = adjacency.toarray()
    adjacencies = {i: adjacency[i].nonzero()[0].tolist() for i in range(len(ch_names))}
    return adjacencies

def check_volt_of_epochs(epochs, reject, flat):
    epochs_cleaned = epochs.copy().drop_bad(reject=reject, flat=flat, verbose=False)
    count = 0
    rejected_epochs = []
    for i, log in enumerate(epochs_cleaned.drop_log): 
        if log==('NO_DATA',) or log==('MERGE DUPLICATE',):
            continue
        if log:
            rejected_epochs.append(count)
        count += 1
    return rejected_epochs

def calculate_SN_ratio(epochs, option='neighbours', mode='log'):
    epochs_data = epochs.get_data(copy=True, verbose=False)
    n_epochs, n_channels, n_times = epochs_data.shape
    clean_signal = np.zeros((n_epochs, n_channels, n_times))
    
    if option == 'neighbours':
        adjacencies = get_neighbors(epochs)
        for ch in range(n_channels):
            neighbor_indices = adjacencies.get(ch)
            neighbor_indices.remove(ch)
            if neighbor_indices:
                clean_signal[:, ch, :] = np.mean(epochs_data[:, neighbor_indices, :], axis=1)

    elif option == 'mean_epochs':
        clean_signal = np.mean(epochs_data, axis=0, keepdims=True).repeat(n_epochs, axis=0)
    elif option == 'median_epochs':
        clean_signal = np.median(epochs_data, axis=0, keepdims=True).repeat(n_epochs, axis=0)
    else:
        raise ValueError("Invalid option. Choose from 'neighbours', 'mean_epochs', 'median_epochs', 'mean_channels', 'median_channels'.")

    noise_signal = epochs_data - clean_signal

    if mode == 'log':
        snr = 10 * np.log10(np.mean(clean_signal ** 2, axis=2) / np.mean(noise_signal ** 2, axis=2))
    elif mode == 'linear':
        snr = np.mean(clean_signal ** 2, axis=2) / np.mean(noise_signal ** 2, axis=2)
    else:
        raise ValueError("Invalid mode. Choose 'log' or 'linear'.")
    return snr
    
def isolation_forest(epochs, mode='ep'):
    data = epochs.get_data(copy=True)
    n_trials, n_channels, n_times = data.shape
    if mode == 'ep':
        clf = IsolationForest(n_estimators=20, random_state=42, max_features=int(n_channels * n_times / 2))
        preds = clf.fit_predict(data.reshape(n_trials, -1))
        return [i for i, pred in enumerate(preds) if pred == -1]
    elif mode == 'ch':
        clf = IsolationForest(n_estimators=20, random_state=42, max_features=int(n_trials * n_times / 2))
        preds = clf.fit_predict(data.reshape(n_channels, -1))
        return [i for i, pred in enumerate(preds) if pred == -1]
    else:
        return None


def mae(y_true, y_pred):  
    return np.mean(np.abs(y_true - y_pred))

def mape(y_true, y_pred):
    eps=1e-8
    return (np.abs(y_true - y_pred)/(y_true+eps)).mean()*100

def rmse(y_true, y_pred):
    return np.sqrt(np.mean((y_true - y_pred)**2))

def rmspe(y_true, y_pred):
    eps=1e-8
    return np.sqrt(np.mean(((y_true - y_pred)/(y_true+eps))**2))*100

def cos_distance(y_true, y_pred):
    cos_alpha = np.dot(y_true, y_pred)/(np.linalg.norm(y_true)*np.linalg.norm(y_pred))
    return cos_alpha

def compute_psd(signal, sfreq):
    freqs = np.fft.rfftfreq(len(signal), d=1./sfreq)
    spectrum = np.abs(np.fft.rfft(signal))**2  
    #  PSD      
    psd = spectrum / (len(signal) * sfreq)
    return freqs, psd

def calculate_statistics(signals, z_score=1.96):
    mean = np.mean(signals, axis=(0, 1))
    std = np.std(signals, axis=0)
    se = np.mean(std / np.sqrt(signals.shape[0]), axis=0)
    ci_95 = z_score * se  # 95%  
    return mean, ci_95

def rythm_metric(epochs, picks):
    brain_bands = {
        'delta': (1, 4),
        'theta': (4, 8),
        'alpha': (8, 13),
        'beta': (14, 30),
        'gamma': (30, 90)
        }
    sfreq = epochs.info['sfreq']
 
    def filter_band(epochs, band):
        return epochs.copy().filter(band[0], band[1])
 
    def get_signals_and_stats(epochs, picks, band):
        filtered_epochs = filter_band(epochs, band)
        signals = filtered_epochs.pick_channels(picks).get_data(copy=True) * 1e6
        mean_signal = signals.mean(axis=(0, 1))
        mean, ci = calculate_statistics(signals)
        freqs, spectrum = compute_psd(mean_signal, sfreq)
        return mean_signal, mean, ci, freqs, spectrum
 
    results = {}
    for band_name, band in brain_bands.items():
        results[band_name] = get_signals_and_stats(epochs, picks, band)

    times = epochs.times

    fig, axs = plt.subplots(5, 2, figsize=(15, 15), sharex='col')

    colors = {
        'delta': 'purple',
        'theta': 'red',
        'alpha': 'blue',
        'beta': 'orange',
        'gamma': 'green'
    }

    for i, (band_name, (signal, mean_signal, ci, freqs, spectrum)) in enumerate(results.items()):
        label = f'{band_name.capitalize()} ({brain_bands[band_name][0]}-{brain_bands[band_name][1]} Hz)'
        color = colors[band_name]
        
        axs[i, 0].plot(times, signal, label=f'{label} SE={np.mean(ci):.2f}$\mu$V', color=color)
        axs[i, 0].set_ylabel('Amplitude ($\mu$V)')
        axs[i, 0].set_title(label)
        axs[i, 0].set_xlabel('Time (s)')
        axs[i, 0].set_xlim(-0.2, 1.0)
        axs[i, 0].fill_between(times, mean_signal - ci, mean_signal + ci, color=color, alpha=0.3)
        axs[i, 0].axvline(x=0, color='black', linestyle='--')
        
        axs[i, 1].plot(freqs, spectrum, label=f'{label} Spectrum', color=color)
        axs[i, 1].set_ylabel('Power/Hz ($\mu V^2$/Hz)')
        axs[i, 1].set_title(f'{label} Spectrum')
        axs[i, 1].set_xlabel('Frequency (Hz)')
        axs[i, 1].set_xlim(0, 90)

    for ax in axs.flat:
        ax.legend()
        ax.grid(True)

    plt.tight_layout()
    plt.show()
    delta_signal = results['delta'][0]
    theta_signal = results['theta'][0]
    alpha_signal = results['alpha'][0]
    beta_signal = results['beta'][0]
    gamma_signal = results['gamma'][0]

    return delta_signal, theta_signal, alpha_signal, beta_signal, gamma_signal

def get_epochs(raw, chns):
    epochs = make_RiTi_events(raw)
    mne.viz.plot_epochs_image(epochs, picks=chns, combine="mean", vmin=-30, vmax=30)
    return epochs

def mahal(x, y, cov=None, reg=1e-20):
   cov = cov if cov else np.cov(np.vstack([x, y]), rowvar=False)
   invcov = np.linalg.pinv(cov, rcond=reg)
   return np.sqrt((x-y)@invcov@(x-y).T)

def mean_amplitude_ratio(signal1, signal2):
    return (np.mean(np.abs(signal1)) - np.mean(np.abs(signal2))) / np.mean(np.abs(signal2)) * 100


def tetta_betta_ratio(tetta, beta):
    return (tetta**2).mean()/(beta**2).mean()

def compare_epochs(inst1, inst2, picks):
    if isinstance(inst1, mne.io.brainvision.brainvision.RawBrainVision) and isinstance(inst2, mne.io.brainvision.brainvision.RawBrainVision):
        epochs1 = get_epochs(inst1, picks)
        epochs2 = get_epochs(inst2, picks)
    else:
        epochs1 = inst1
        epochs2 = inst2
    delta_signals_1, theta_signals_1, alpha_signals_1, beta_signals_1, gamma_signals_1 = rythm_metric(epochs1, picks)
    delta_signals_2, theta_signals_2, alpha_signals_2, beta_signals_2, gamma_signals_2 = rythm_metric(epochs2, picks)
    results = {'Delta':{}, 'Theta':{}, 'Alpha':{}, 'Beta':{}, 'Gamma':{}}
    for i, (signal1, signal2, band) in enumerate([(delta_signals_1, delta_signals_2, 'Delta'), (theta_signals_1, theta_signals_2, 'Theta'),
                                            (alpha_signals_1, alpha_signals_2, 'Alpha'), (beta_signals_1, beta_signals_2, 'Beta'), 
                                            (gamma_signals_1, gamma_signals_2, 'Gamma')]):
        results[band]['MAE'] = mae(signal1, signal2)
        results[band]['MAPE(%)'] = mape(signal1, signal2)
        results[band]['RMSE'] = rmse(signal1, signal2)
        results[band]['RMSPE(%)'] = rmspe(signal1, signal2)
        results[band]['CosA'] = cos_distance(signal1, signal2)
        results[band]['Mahal'] = mahal(signal1, signal2)
        results[band]['Mean-A'] = mean_amplitude_ratio(signal1, signal2)

    df_results = pd.DataFrame(results)
    
    print(df_results, f'\nAplitude~ V*1e-6')
    print('Tetta Beta Ratio', tetta_betta_ratio(theta_signals_1, beta_signals_1)/tetta_betta_ratio(theta_signals_2, beta_signals_2))

def cross_correlation_score(data1, data2, sf):
    data1 = np.mean(data1, axis=0)
    data2 = np.mean(data2, axis=0)
    corr = correlate(data1, data2)
    lags = np.arange(-len(data1) + 1, len(data2))
    lags_in_seconds = lags / sf
    plt.plot(lags_in_seconds, corr)
    plt.title(f'-')
    plt.xlabel(' (s)')
    plt.ylabel('')
    plt.show()


def test_chi_square_mormal(epochs):
    epochs_data = epochs.get_data(copy=True)
    results = []
    for epoch in epochs_data:
        pvals = []
        for channel in epoch:
            mean, std = norm.fit(channel)
            observed_freqs, bins = np.histogram(channel, bins='auto')
            expected_freqs = norm.cdf(bins[1:], mean, std) - norm.cdf(bins[:-1], mean, std)
            expected_freqs *= observed_freqs.sum() / expected_freqs.sum() 
            chi_stat, p_chi_square = chisquare(f_obs=observed_freqs, f_exp=expected_freqs)
            pvals.append(p_chi_square)
        results.append(pvals)
    results = np.array(results)
    cmap = plt.cm.RdYlGn_r 
    fig = plt.figure(figsize=(64, 8))
 
    cax2 = plt.imshow(results.T, cmap=cmap, vmin=0, vmax=0.2, aspect=9.0)
    fig.colorbar(cax2)
    plt.yticks(ticks=np.arange(results.shape[1]), labels=epochs.ch_names, fontsize=12)
    plt.title("Chi-square Test P-values", fontsize=14)
    plt.xlabel('Epochs', fontsize=12)
    plt.ylabel('Channels', fontsize=12)
    plt.show()

def test_Shapiro_mormal(epochs):
    epochs_data = epochs.get_data(copy=True)  
    results = []
    for epoch in epochs_data:
        pvals = []
        for channel in epoch:
            stat, p_dagostino = shapiro(channel)
            pvals.append(p_dagostino)
        results.append(pvals)
    results = np.array(results)
    cmap = plt.cm.RdYlGn_r  
    fig = plt.figure(figsize=(64, 8))
    cax2 = plt.imshow(results.T, cmap=cmap, vmin=0, vmax=0.2, aspect=9.0)
    fig.colorbar(cax2)
    plt.yticks(ticks=np.arange(results.shape[1]), labels=epochs.ch_names, fontsize=12)
    plt.title("D'Shapiro Test P-values", fontsize=14)
    plt.xlabel('Epochs', fontsize=12)
    plt.ylabel('Channels', fontsize=12)
    plt.show()

def test_Dagostino_mormal(epochs):
    epochs_data = epochs.get_data(copy=True)
    results = []
    for epoch in epochs_data:
        pvals = []
        for channel in epoch:
            stat, p_dagostino = normaltest(channel)
            pvals.append(p_dagostino)
        results.append(pvals)
    results = np.array(results)
    cmap = plt.cm.RdYlGn_r 
    fig = plt.figure(figsize=(64, 8))
    cax2 = plt.imshow(results.T, cmap=cmap, vmin=0, vmax=0.2, aspect=9.0)
    fig.colorbar(cax2)
    plt.yticks(ticks=np.arange(results.shape[1]), labels=epochs.ch_names, fontsize=12)
    plt.title("D'Agostino-K^2 Test P-values", fontsize=14)
    plt.xlabel('Epochs', fontsize=12)
    plt.ylabel('Channels', fontsize=12)
    plt.show()



# Copyright 2025 Sear Gamemode
import mne 
import torch
import numpy as np 
import torch.nn as nn
import torch.optim as optim
import torch.nn.functional as F
from .metrics import calculate_SN_ratio

class Encoder(nn.Module):
    def __init__(self, input_dim, latent_dim):
        super(Encoder, self).__init__()
        self.fc1 = nn.Linear(input_dim, 128)
        self.fc2 = nn.Linear(128, 64)
        self.fc_mu = nn.Linear(64, latent_dim)  #  
        self.fc_logvar = nn.Linear(64, latent_dim)  #  

    def forward(self, x):
        x = F.relu(self.fc1(x))
        x = F.relu(self.fc2(x))
        mu = self.fc_mu(x)
        logvar = self.fc_logvar(x)
        return mu, logvar

class Decoder(nn.Module):
    def __init__(self, latent_dim, output_dim):
        super(Decoder, self).__init__()
        self.fc1 = nn.Linear(latent_dim, 64)
        self.fc2 = nn.Linear(64, 128)
        self.fc3 = nn.Linear(128, output_dim)

    def forward(self, z):
        z = F.relu(self.fc1(z))
        z = F.relu(self.fc2(z))
        z = self.fc3(z)
        return z

class VAE(nn.Module):
    def __init__(self, input_dim, latent_dim):
        super(VAE, self).__init__()
        self.encoder = Encoder(input_dim, latent_dim)
        self.decoder = Decoder(latent_dim, input_dim)

    def reparameterize(self, mu, logvar):
        std = torch.exp(0.5 * logvar)
        eps = torch.randn_like(std)
        return mu + eps * std

    def forward(self, x):
        mu, logvar = self.encoder(x)
        z = self.reparameterize(mu, logvar)
        reconstructed_x = self.decoder(z)
        return reconstructed_x, mu, logvar, z

def loss_function(reconstructed_x, x, mu, logvar, z, alpha=0, beta=0.5):
    recon_loss = F.mse_loss(reconstructed_x, x, reduction='mean')
    kld_loss = -0.5 * torch.sum(1 + logvar - mu.pow(2) - logvar.exp())
    infomax_loss = alpha * (z.pow(2).mean(dim=0) - 1).abs().sum()
    mae_loss = F.l1_loss(reconstructed_x, x, reduction='mean')
    return recon_loss + kld_loss + infomax_loss + beta*mae_loss

class Feature_Extractor():
    def __init__(self, epochs, latent_dim=5, lr=1e-3, alpha=1.0):
        # 
        self.alpha=alpha  #   infomax
        self.lr = lr
        input_dim = epochs.get_data().shape[1] * epochs.get_data().shape[2]  #      
        self.latent_dim = latent_dim #   
        #    
        self.vae = VAE(input_dim=input_dim, latent_dim=latent_dim)
        self.optimizer = optim.Adam(self.vae.parameters(), lr=self.lr)
        #  
        epochs_data = torch.tensor(epochs.get_data(), dtype=torch.float32)
        self.epochs_data = epochs_data.view(epochs_data.shape[0], -1)  #   

    def train(self,):
        n_epochs = 100
        for epoch in range(n_epochs):
            self.vae.train()
            self.optimizer.zero_grad()
            reconstructed_x, mu, logvar, z = self.vae(self.epochs_data)
            loss = loss_function(reconstructed_x, self.epochs_data, mu, logvar, z, alpha=self.alpha)
            loss.backward()
            self.optimizer.step()
            if (epoch + 1) % 10 == 0:
                print(f" [{epoch+1}/{n_epochs}], : {loss.item():.4f}")
    
    def extract(self, epochs):
        epochs_data = torch.tensor(epochs.get_data(), dtype=torch.float32)
        epochs_data = epochs_data.view(epochs_data.shape[0], -1) 
        self.vae.eval()
        with torch.no_grad():
            mu, logvar = self.vae.encoder(epochs_data)
            z = self.vae.reparameterize(mu, logvar)  
        return z

def get_VAE_features(epochs, vae):
    z = vae.extract(epochs)
    return z

def snr(epochs):
    snr_matrix = calculate_SN_ratio(epochs)
    return {'SNR': snr_matrix}

def mean(epochs, mode='zero'):
    data = epochs.get_data()
    if mode == 'zero':
        tmin = epochs.tmin
        sfreq = epochs.info['sfreq']
        result = data[:, :, int(abs(tmin*sfreq)):].mean(axis=2)
    else:
        result = data.mean(axis=2)
    return {"Mean": result}

def std(epochs, mode='zero'):
    data = epochs.get_data()
    if mode == 'zero':
        tmin = epochs.tmin
        sfreq = epochs.info['sfreq']
        result = data[:, :, int(abs(tmin*sfreq)):].std(axis=2)
    else:
        result = data.std(axis=2)
    return {"Std": result}

def median(epochs, mode='zero'):
    data = epochs.get_data()
    if mode == 'zero':
        tmin = epochs.tmin
        sfreq = epochs.info['sfreq']
        result = np.median(data[:, :, int(abs(tmin*sfreq)):], axis=2)
    else:
        result = data.std(axis=2)
    return {"Std": result}

def bands_spectrum_power(epochs):
    results = {}
    bands =  {'delta': (1, 4), 'theta': (4, 8), 'alpha': (8, 13), 'beta': (14, 30), 'gamma': (30, 90)}
    data = epochs.get_data()
    sfreq = epochs.info['sfreq']
    for band_name, (fmin, fmax) in bands.items():
        psds, _ = mne.time_frequency.psd_array_multitaper(data, sfreq=sfreq, fmin=fmin, fmax=fmax, verbose=False)
        band_power = np.mean(psds, axis=2)
        results[band_name] = band_power
    return results

def derivative(epochs):
    data = epochs.get_data()
    sfreq = epochs.info["sfreq"]
    deriv = (data[1:] - data[:-1])/sfreq
    return {"1-Deriv": deriv}

def standard_error_of_the_mean(epochs):
    data = epochs.get_data()
    n_trials = data.shape[0]
    sem = np.std(data, axis=0) / np.sqrt(n_trials)
    result = {'SEM':sem}
    return result

def latency_and_peaks(epochs, component):
    peak_windows = {
        'P1': (0.08, 0.13),
        'N1': (0.10, 0.15),
        'P2': (0.15, 0.25),
        'N2': (0.20, 0.30),
        'P3': (0.30, 0.50)
    }
    if component not in peak_windows:
        raise ValueError(f" : {component}.  : {list(peak_windows.keys())}")
    
    tmin, tmax = peak_windows[component]
    evoked = epochs.average()
    
    results = {
        f"{component}_latency": [],
        f"{component}_amplitude": []
    }
    for ch_name in evoked.ch_names:
        latency, amplitude = evoked.get_peak(
            tmin=tmin, tmax=tmax, mode='pos' if 'P' in component else 'neg', 
            ch_type='eeg', picks=[ch_name], return_amplitude=True
        )
        
    results[f"{component}_latency"].append(latency)
    results[f"{component}_amplitude"].append(amplitude)

    return results


def feature_selector(epochs, list_features=[]):
    values = {}
    for metric in list_features:
        values[metric.__name__] = metric(epochs)
    return values





# Copyright 2025 Sear Gamemode
import re 
from datetime import datetime
import os 
import mne 
import numpy as np 


def preprocessing_events(raw, scenarious='RiTi'):
    if scenarious == 'RiTi':
        sequence_file = 'RiTi_seq.txt'
        seq_list = []
        transform = {'Stimulus/s115':1, 'Stimulus/s120':4, 'Stimulus/s130':2, 'Stimulus/s160':3, 'Stimulus/s240':5}
        reverse_transform = dict(zip(transform.values(), transform.keys()))
        with open(sequence_file, 'r') as file:
            for line in file:
                seq_list.append(reverse_transform[int(line)])
        events, event_id = mne.events_from_annotations(raw, verbose=False)
        reverse_event_id = dict(zip(event_id.values(), event_id.keys()))
        new_events = verificate_events(raw, seq_list)
        filtered_seq = [reverse_event_id[new_events[j, 2]] for j in range(len(new_events))]
        subseq, match_idxs, tr = is_subsequence(filtered_seq, seq_list)
        repo_data = {   
                    'Initial_Events': len(events),
                    'Filtered_Events': len(new_events),
                    'Expected_Events': len(seq_list),
                    'SubSeq_Flag': subseq, 
                    'Match_Indexes': match_idxs,
                    'Transitions': tr
                    }
        return new_events, reverse_event_id, repo_data
    else:
        print('Unknown scenarious!')
        return None
    

def get_meta(file_name):
    file_pattern = re.compile(
        r'INP(?P<id>\d{4})_v1\.(?P<visit_num>\d)_(?P<experiment>[^_]+)_'
        r'(?P<operator_code>[A-Za-z0-9_]+)_(?P<date>\d{2}\.\d{2}\.\d{2,4})'
        r'\.(?P<format>vhdr|vmrk|eeg)')
    match = file_pattern.match(file_name)
    return match.groupdict() if match else None


def get_brainvision_files(vhdr_path):
    eeg_fname = None
    vmrk_fname = None
    with open(vhdr_path, 'r', encoding='utf-8') as f:
        lines = f.readlines()
    for line in lines:
        line = line.strip()
        if line.startswith('DataFile='):
            eeg_fname = line.split('=')[1].strip()
        elif line.startswith('MarkerFile='):
            vmrk_fname = line.split('=')[1].strip()
    return (eeg_fname, vmrk_fname)


def get_file_info(file_path, elc_path):
    raw = mne.io.read_raw(file_path, preload=False, verbose=False)
    other_files = ()
    if os.path.splitext(file_path)[1]=='.vhdr':
        other_files = get_brainvision_files(file_path)
    eeg_files = list(map(os.path.basename, (file_path,) + other_files))
    meta = get_meta(eeg_files[0])
    meas_date = raw.info['meas_date']
    data = {
            "id": meta.get('id', 'N/A'),
            "visit_num": meta.get('visit_num', 'N/A'),
            "scenario": meta.get('experiment', 'N/A'),
            "operator_code": ', '.join(meta.get('operator_code', 'N/A').split('_')),
            "eeg_files": ' '.join(eeg_files),
            "elc_file": os.path.basename(elc_path) if elc_path else 'Not found',
            "raw_type": type(raw).__name__,
            "duration": f"{raw.times[-1] - raw.times[0]:.2f}",
            "nchan": raw.info['nchan'],
            "electrodes_of_montages" : [raw.info['ch_names'][i] for i in mne.pick_types(raw.info, eeg=True)],
            "ecg_channel": not (set([ch for ch in raw.ch_names if 'BIP' in ch]) is {}),
            "eog_channel": not (set([ch for ch in raw.ch_names if 'EOG' in ch]) is {}),
            "sfreq": raw.info['sfreq'],
            "measurement_datetime": meas_date.strftime('%Y-%m-%d %H:%M:%S') if isinstance(meas_date, datetime) else 'Unknown',
            "highpass": raw.info['highpass'],
            "lowpass": raw.info['lowpass'],
            "custom_ref_applied": raw.info.get('custom_ref_applied', 'Not available'),
    }
    return data


def is_subsequence(subseq, seq):
    subseq_index = 0
    subseq_len = len(subseq)
    flag = 0
    transitions = 0
    for elem in seq:
        copy_flag = flag
        if subseq_index < subseq_len and elem == subseq[subseq_index]:
            subseq_index += 1
            flag = 1
        else:
            flag = 0
        if flag != copy_flag:
            transitions += 1
    return subseq_index == subseq_len, subseq_index, transitions


def verificate_events(raw, seq_list, shift=1.85, var=0.35):
    events, event_id = mne.events_from_annotations(raw, verbose=False)
    reverse_event_id = dict(zip(event_id.values(), event_id.keys()))
    tar_idx = 0
    new_events = []
    n_events = len(events)
    n_seq = len(seq_list)
    events_times = events[:, 0]/raw.info['sfreq']
    events_seq = np.array([reverse_event_id[events[i, 2]] for i in range(len(events))])
    current_time = events_times[0] - shift

    while not (tar_idx > n_seq - 1 or current_time > events_times[-1]):
        
        next_time_l = current_time + shift - var
        next_time_r = current_time + shift + var

        mask = (events_times >= next_time_l) & (events_times <= next_time_r)
        filtered_seq = events_seq[mask]
        filtered_events = events[mask]
        filtered_times = events_times[mask]
        if filtered_seq.size > 0:
            unique_elements = np.unique(filtered_seq)
            if len(unique_elements) == 1:
                new_events.append(filtered_events[0])
                tar_idx += 1
                current_time = filtered_times[0]
            else:
                matches = seq_list[tar_idx] == filtered_seq
                if matches.all():
                    new_events.append(filtered_events[matches[0]])
                    tar_idx += 1
                    current_time = filtered_times[matches[0]]
                else:    
                    tar_idx += 1
                    current_time += shift
        else:
            tar_idx += 1
            current_time += var
    
    return np.array(new_events)
